import pathlib

import joblib
import numcodecs
import numpy as np
import pandas as pd
import pyBigWig
import pyranges as pr
import zarr
from numba import jit


@jit(nopython=True)
def _get_chr_zarr_data(region_data, chr_size):
    zarr_data = np.zeros((chr_size,), dtype="int32")
    for start, end, value in region_data:
        zarr_data[start:end] = value
    return zarr_data


def coverage_to_zarr(coverage_path, zarr_path, chrom_sizes_path, chunk_size=20000000):
    """
    Convert a Tn5-currected cut site coverage file to a zarr file.

    Parameters
    ----------
    coverage_path : str
        Path to the Tn5-currected cut site coverage file generated by Genrich.
    zarr_path : str
        Path to the output zarr file.
    chrom_sizes_path : str
        Path to the chromosome sizes file.
    chunk_size : int
        Chunk size for the zarr file.

    Returns
    -------
    None
    """
    chrom_sizes = pd.read_csv(chrom_sizes_path, index_col=0, sep="\t", header=None).squeeze().to_dict()

    coverage = pd.read_csv(
        coverage_path,
        comment="#",
        sep="\t",
        usecols=[0, 1, 2, 3],
        dtype={"chr": "str", "start": "int32", "end": "int32", "experimental": "int32"},
    )
    coverage = coverage[coverage.iloc[:, 3] > 0].copy()

    pathlib.Path(zarr_path).mkdir(exist_ok=True, parents=True)
    root = zarr.group(store=zarr_path, overwrite=True)  # init zarr group
    zarr_compressor = numcodecs.Blosc(cname="zstd", clevel=3, shuffle=numcodecs.Blosc.SHUFFLE)  # Setup compressor

    for chr, chr_df in coverage.groupby("chr"):
        try:
            chr_size = chrom_sizes[chr]
        except KeyError:
            continue

        region_data = chr_df.iloc[:, 1:].values
        zarr_data = _get_chr_zarr_data(region_data=region_data, chr_size=chr_size)
        root.create_dataset(
            f"chrs/{chr}",
            data=zarr_data,
            chunks=chunk_size,
            compressor=zarr_compressor,
        )

    # final success flag
    pathlib.Path(f"{zarr_path}/.success").touch()
    return


def normalize_zarr(zarr_path, peak_path, chrom_sizes_path, output_zarr_path=None, chunk_size=20000000):
    """
    Normalize the raw cut site coverage zarr file.

    The normalization is done by subtracting the mean signal in non-peak regions and dividing by the 75th percentile of the signal in peak regions.
    Peak and non-peak regions are defined by a bed file, typically those ensambled peak list from ENCODE or atlas projects.

    Parameters
    ----------
    zarr_path : str
        Path to the input zarr file.
    peak_path : str
        Path to the peak regions bed file.
    chrom_sizes_path : str
        Path to the chromosome sizes file.
    output_zarr_path : str
        Path to the output zarr file. If None, only calculate the normalization factors and save them to the input zarr file directory at "{zar_path}/.norm".
    chunk_size : int
        Chunk size for the zarr file.

    Returns
    -------
    None
    """
    # Read peak and non-peak regions (bed file)
    genome_bed = pd.read_csv(chrom_sizes_path, header=None, sep="\t", names=["Chromosome", "End"])
    genome_bed["Start"] = 0
    genome_bed = pr.PyRanges(genome_bed[["Chromosome", "Start", "End"]])

    peaks = pr.read_bed(peak_path)
    non_peaks = genome_bed.subtract(peaks)
    peaks = peaks.as_df()
    non_peaks = non_peaks.as_df()

    # Read zarr track file to numpy array
    zarr_track_dir = zarr.open(zarr_path, mode="r")

    # Calculate the mean signal in non-peak regions
    non_peak_signals = []
    for chr_name, track in zarr_track_dir["chrs"].items():
        track = np.array(track)
        non_peaks_chr = non_peaks[non_peaks["Chromosome"] == chr_name]
        starts = non_peaks_chr.iloc[:, 1]
        ends = non_peaks_chr.iloc[:, 2]
        non_peak_signals_chr = [track[start:end].mean() for start, end in zip(starts, ends)]
        non_peak_signals += non_peak_signals_chr
    non_peak_mean = np.mean(non_peak_signals)

    # Calculate the 75th percentile signal in peak regions
    peak_signals = []
    for chr_name, track in zarr_track_dir["chrs"].items():
        track = np.array(track)
        peaks_chr = peaks[peaks["Chromosome"] == chr_name]
        starts = peaks_chr.iloc[:, 1]
        ends = peaks_chr.iloc[:, 2]
        peak_signals_chr = [track[start:end].mean() for start, end in zip(starts, ends)]
        peak_signals += peak_signals_chr
    peak_percentile = np.percentile(peak_signals, np.arange(99, 99.9, 0.1)).mean()

    joblib.dump(
        {"background_offset": non_peak_mean, "peak_scale": peak_percentile},
        f"{zarr_path}/.norm",
    )

    if output_zarr_path is not None:
        # Initialize the new zarr track similar to the zarr_track_dir
        new_zarr_track_dir = zarr.open_group(output_zarr_path, mode="w")
        new_zarr_track = new_zarr_track_dir.create_group("chrs")
        for chr_name, track in zarr_track_dir["chrs"].items():
            # Subtract the median signal in non-peak regions
            # Divide by the 75th percentile of the signal in peak regions
            new_data = (track[:] - non_peak_mean) / peak_percentile
            new_zarr_track.create_dataset(
                chr_name,
                data=new_data,
                chunks=chunk_size,
                shape=track.shape,
                dtype="float32",
            )


def zarr_to_bigwig(zarr_path, bigwig_path):
    """
    Convert a zarr file to a bigWig file.

    Parameters
    ----------
    zarr_path : str
        Path to the input zarr file.
    bigwig_path : str
        Path to the output bigWig file.

    Returns
    -------
    None
    """
    # Open the existing zarr group
    root = zarr.open_group(store=zarr_path, mode="r")
    chromsizes = {name: array.shape[0] for name, array in root["chrs"].items()}

    # Define headers
    header = list(chromsizes.items())

    # Open a new bigWig file for writing
    with pyBigWig.open(str(bigwig_path), "w") as bw:
        # Add the header
        bw.addHeader(header, maxZooms=10)

        # Loop over all chromosomes in the zarr group
        for chr_name, _ in header:
            print(f"Processing {chr_name}")
            # Load the dense array for this chromosome
            array = root["chrs"][chr_name][:]

            diff = np.diff(array, prepend=array[0] - 1, append=array[-1] + 1)
            change_points = np.where(diff)[0]

            starts = change_points[:-1]
            ends = change_points[1:]
            values = array[starts].astype(float)
            chroms = np.repeat(chr_name, len(starts))

            bw.addEntries(chroms, starts, ends, values=values)



import pathlib

import anndata
import numpy as np
import pandas as pd
import xarray as xr
import zarr
from numcodecs import Blosc


def _chrom_sizes_to_chorm_offset_table(chrom_sizes):
    # chrom offset table
    chrom_sizes = chrom_sizes.sort_index()
    chrom_offset_table = pd.DataFrame({"size": chrom_sizes, "global_end": chrom_sizes.cumsum()})
    chrom_offset_table["global_start"] = chrom_offset_table["global_end"] - chrom_sizes
    chrom_offset_table = chrom_offset_table[["global_start", "global_end", "size"]].copy()
    chrom_offset_table.index.name = "chrom"
    chrom_offset_table.columns.name = "offset_value"
    return chrom_offset_table


def _get_bedpe_fragments(
    bedpe_chunk,
    chr1,
    start1,
    chr2,
    end2,
    mapq_col,
    min_mapq,
    strand1,
    strand2,
    barcode_col,
    use_chroms,
    adjust_tn5,
):
    flags = []
    if use_chroms is not None:
        flags.append(bedpe_chunk[chr1].isin(use_chroms))
    if chr2 is not None:
        flags.append(bedpe_chunk[chr1] == bedpe_chunk[chr2])
    if mapq_col is not None:
        flags.append(bedpe_chunk[mapq_col] >= min_mapq)
    if strand1 is not None and strand2 is not None:
        flags.append(bedpe_chunk[strand1] != bedpe_chunk[strand2])

    bedpe_chunk = bedpe_chunk.iloc[np.all(flags, axis=0)]

    if adjust_tn5:
        r1_is_pos_strand = bedpe_chunk[strand1] == "+"
        r1_start_adj = np.where(r1_is_pos_strand, -5, 4)
        bedpe_chunk.iloc[:, start1] += r1_start_adj
        r2_end_adj = np.where(r1_is_pos_strand, 4, -5)
        bedpe_chunk.iloc[:, end2] += r2_end_adj

    frag_bed = bedpe_chunk.iloc[:, [chr1, start1, end2, barcode_col]].copy()
    return frag_bed


def fragments_to_cutsite_zarr(
    fragments_path,
    chrom_sizes_path,
    output_zarr_path,
    barcode_prefix=None,
    chrom_col=0,
    start_col=1,
    end_col=2,
    barcode_col=3,
    chrom2_col=None,
    end2_col=None,
    strand_col=None,
    strand2_col=None,
    mapq_col=None,
    min_mapq=30,
    format="bed",
    adjust_tn5=False,
    chunk_size=20000000,
):
    """
    Extracts the unique cut sizes for each cell barcode and stores them in a zarr store.

    Parameters
    ----------
    fragments_path : str
        Path to the fragments file. The file should be a tab-separated file with columns
        for chromosome, start position, end position, and barcode.
    chrom_sizes_path : str
        Path to the chromosome sizes file. The file should be a tab-separated file with
        two columns: chromosome name and size.
    output_zarr_path : str
        Path to the output zarr store.
    barcode_prefix : str, optional
        Prefix to add to the barcode to make it globally unique across multiple datasets. Default is None.
    chrom_col : int, optional
        Column index for the chromosome in the fragments file. Default is 0.
    start_col : int, optional
        Column index for the start position in the fragments file. Default is 1.
    end_col : int, optional
        Column index for the end position in the fragments file. Default is 2.
    barcode_col : int, optional
        Column index for the barcode in the fragments file. Default is 3 for bed format and 6 for bedpe format.
    chrom2_col : int, optional
        Column index for the second chromosome in the bedpe fragments file. If format is bedpe, this will be set to 3.
        Default is None.
    end2_col : int, optional
        Column index for the second end position in the bedpe fragments file. If format is bedpe, this will be set to 5.
        Default is None.
    strand_col : int, optional
        Column index for the strand in the bedpe fragments file. If format is bedpe, this will be set to 8. Default is None.
    strand2_col : int, optional
        Column index for the second strand in the bedpe fragments file. If format is bedpe, this will be set to 9.
        Default is None.
    mapq_col : int, optional
        Column index for the mapping quality in the bedpe fragments file. If format is bedpe, this will be set to 7.
        Default is None.
    min_mapq : int, optional
        Minimum mapping quality for the fragments. Default is 30.
    format : str, optional
        Format of the fragments file. Default is "bed". Other option is "bedpe".
    adjust_tn5 : bool, optional
        Adjust the TN5 positions for the fragments. Default is False.
    chunk_size : int, optional
        Chunk size for the zarr store. Default is 20,000,000.

    Returns
    -------
    None
    """
    # check if zarr store already exists and there is a success flag
    success_flag = pathlib.Path(output_zarr_path) / ".success"
    if pathlib.Path(output_zarr_path).exists():
        if success_flag.exists():
            print(f"Zarr store already exists at {output_zarr_path}")
            return

    # set default column indices for bedpe
    if format == "bedpe":
        chrom2_col = 3
        end2_col = 5
        barcode_col = 6
        mapq_col = 7
        strand_col = 8
        strand2_col = 9
        print("Using bedpe format from BEDTOOLS")
        if adjust_tn5:
            print("Adjusting TN5 positions")

    if chrom2_col is not None:
        bedpe = True
    else:
        bedpe = False

    zarr.storage.default_compressor = Blosc()

    chrom_sizes = pd.read_csv(chrom_sizes_path, header=None, sep="\t", index_col=0).squeeze()
    chrom_offset_table = _chrom_sizes_to_chorm_offset_table(chrom_sizes)

    # process fragments
    frag_chunks = pd.read_csv(fragments_path, chunksize=1000000, header=None, sep="\t")
    cur_barcode_id = 0
    barcode_id_map = {}
    records = {}
    for chunk in frag_chunks:
        if bedpe:
            # process bedpe and adjust for TN5, return the fragment bed
            chunk = _get_bedpe_fragments(
                bedpe_chunk=chunk,
                chr1=chrom_col,
                start1=start_col,
                chr2=chrom2_col,
                end2=end2_col,
                mapq_col=mapq_col,
                min_mapq=min_mapq,
                strand1=strand_col,
                strand2=strand2_col,
                barcode_col=barcode_col,
                use_chroms=chrom_offset_table.index,
                adjust_tn5=adjust_tn5,
            )

        for barcode, barcode_df in chunk.groupby(barcode_col):
            if barcode in barcode_id_map:
                barcode_idx = barcode_id_map[barcode]
                exist = True
            else:
                barcode_idx = cur_barcode_id
                cur_barcode_id += 1
                barcode_id_map[barcode] = barcode_idx
                exist = False

            use_frags_df = barcode_df[barcode_df[chrom_col].isin(chrom_offset_table.index)].iloc[
                :, [chrom_col, start_col, end_col]
            ]
            global_cutsites = (
                use_frags_df.iloc[:, 1:]
                + use_frags_df.iloc[:, 0].map(chrom_offset_table["global_start"]).values[:, None]
            )
            global_cutsites = np.sort(np.unique(global_cutsites.values.ravel()))

            if exist:
                cur_cutsites = records[barcode_idx]
                new_cutsites = np.sort(np.unique(np.concatenate([cur_cutsites, global_cutsites])))
                records[barcode_idx] = new_cutsites

            else:
                records[barcode_idx] = global_cutsites

    # prepare an empty zarr store
    total_cutsites_count = sum(v.size for v in records.values())
    final_array = xr.DataArray(
        np.zeros((total_cutsites_count, 2), dtype="uint32"),
        dims=["site", "value"],
        coords={"value": ["barcode", "global_pos"]},
    )
    final_array = final_array.chunk({"site": chunk_size, "value": "1"})
    final_ds = xr.Dataset({"cutsite": final_array})
    # apply barcode prefix if provided
    if barcode_prefix is not None:
        barcode_id_map = {f"{barcode_prefix}:{k}": v for k, v in barcode_id_map.items()}
    final_ds["barcode_map"] = xr.DataArray(pd.Series(barcode_id_map), dims="barcode")
    final_ds["chrom_offset"] = xr.DataArray(chrom_offset_table, dims=["chrom", "offset_value"])
    for k in list(final_ds.coords.keys()):
        v = final_ds.coords[k]
        final_ds.coords[k] = v.astype(str)
        final_ds.coords[k].encoding = {"chunks": (v.size,)}
    final_ds.to_zarr(output_zarr_path, mode="w")

    # write the cutsites in large chunks
    final_array = zarr.open_array(f"{output_zarr_path}/cutsite/")
    mat_records = []
    temp_site_counts = chunk_size
    cur_start = 0
    for barcode, cutsites in records.items():
        mat = np.zeros((cutsites.size, 2), dtype="uint32")
        mat[:, 0] = barcode
        mat[:, 1] = cutsites
        mat_records.append(mat)
        temp_site_counts -= cutsites.size

        if (temp_site_counts < 0) and len(mat_records) != 0:
            mat_records = np.concatenate(mat_records)
            cur_end = cur_start + mat_records.shape[0]
            final_array[cur_start:cur_end, :] = mat_records

            cur_start = cur_end
            mat_records = []
            temp_site_counts = chunk_size

    # final chunk
    if len(mat_records) != 0:
        mat_records = np.concatenate(mat_records)
        cur_end = cur_start + mat_records.shape[0]
        final_array[cur_start:cur_end, :] = mat_records

    # write a success flag to the zarr store
    success_flag.touch()
    return


def adata_cutsites_to_zarr(adata_path, output_zarr_path, barcode_prefix=None, chunk_size=20000000):
    """
    Converts the insertion matrix (sparse format) from SnapATAC2 anndata object into a zarr store.

    Parameters
    ----------
    adata_path : str
        Path to the anndata object.
    output_zarr_path : str
        Path to the output zarr store.
    barcode_prefix : str, optional
        Prefix to add to the barcode to make it globally unique across multiple datasets. Default is None.
    chunk_size : int, optional
        Chunk size for the zarr store. Default is 20,000,000.

    Returns
    -------
    None
    """
    success_flag = pathlib.Path(output_zarr_path) / ".success"
    if pathlib.Path(output_zarr_path).exists():
        if success_flag.exists():
            print(f"Zarr store already exists at {output_zarr_path}")
            return

    adata = anndata.read_h5ad(adata_path)

    assert "insertion" in adata.obsm_keys(), "insertion matrix not found in adata.obsm"
    assert "reference_sequences" in adata.uns_keys(), "reference_sequences not found in adata.uns"

    # prepare chrom offset table directly from adata.uns
    chrom_sizes = adata.uns["reference_sequences"].set_index("reference_seq_name").squeeze()
    chrom_offset_table = _chrom_sizes_to_chorm_offset_table(chrom_sizes)

    # turn insertion mat into coo like format for zarr store
    insertions = adata.obsm["insertion"]
    barcode_id_map = {b: i for i, b in enumerate(adata.obs_names)}

    coo = insertions.tocoo()
    zarr_data = np.array([coo.row, coo.col], dtype="uint32").T

    # prepare the zarr store
    final_array = xr.DataArray(
        zarr_data,
        dims=["site", "value"],
        coords={"value": ["barcode", "global_pos"]},
    )
    final_array = final_array.chunk({"site": chunk_size, "value": "1"})
    final_ds = xr.Dataset({"cutsite": final_array})
    # apply barcode prefix if provided
    if barcode_prefix is not None:
        barcode_id_map = {f"{barcode_prefix}:{k}": v for k, v in barcode_id_map.items()}
    final_ds["barcode_map"] = xr.DataArray(pd.Series(barcode_id_map), dims="barcode")
    final_ds["chrom_offset"] = xr.DataArray(chrom_offset_table, dims=["chrom", "offset_value"])
    for k in list(final_ds.coords.keys()):
        v = final_ds.coords[k]
        final_ds.coords[k] = v.astype(str)
        final_ds.coords[k].encoding = {"chunks": (v.size,)}
    final_ds.to_zarr(output_zarr_path, mode="w")

    # write a success flag to the zarr store
    success_flag.touch()
    return



import pytest

import bolero_process


def test_package_has_version():
    assert bolero_process.__version__ is not None


@pytest.mark.skip(reason="This decorator should be removed when test passes.")
def test_example():
    assert 1 == 0  # This test is designed to fail.



import numpy as np
import pandas as pd
import pyBigWig
import xarray as xr
import zarr


class CutSitesZarr(xr.Dataset):
    """Class to read cutsite zarr file and provide functionality to convert single cell cutsite data to pseudobulk coverage data."""

    def __init__(self, zarr_path) -> None:
        ds = xr.open_zarr(zarr_path)
        super().__init__(data_vars=ds.data_vars, coords=ds.coords, attrs=ds.attrs)

        self.barcode_to_idx = ds["barcode_map"].to_pandas()
        self.idx_to_barcode = pd.Series({i: b for b, i in self.barcode_to_idx.items()})
        self.chrom_offsets = ds["chrom_offset"].to_pandas()
        self.genome_total_length = self.chrom_offsets["size"].sum()
        self._zarr_compressor = zarr.Blosc(cname="zstd", clevel=3, shuffle=zarr.Blosc.SHUFFLE)

    def get_barcodes_idx(self, barcodes):
        """Get the index of the barcodes in the barcode_map."""
        use_barcodes_idx = self.barcode_to_idx[self.barcode_to_idx.index.isin(barcodes)].values
        return use_barcodes_idx

    def _get_barcodes_sites_count(self, barcodes):
        use_idx = self.get_barcodes_idx(barcodes)
        if use_idx.size == 0:
            return pd.Series(dtype=np.uint32)

        # cutsite chunks
        site_chunks = self["cutsite"].chunks[0]
        chunk_boarders = [0] + list(np.cumsum(site_chunks))

        total_cutsites = []
        for _i in range(len(site_chunks)):
            chunk_start = chunk_boarders[_i]
            chunk_end = chunk_boarders[_i + 1]

            # first, select barcode idx bool array
            use_chunk_sites = np.isin(self["cutsite"][chunk_start:chunk_end, 0], use_idx)
            # second, load actual cutsite global positions with the bool array
            use_chunk_cutsites = self["cutsite"][chunk_start:chunk_end, 1][use_chunk_sites].values
            total_cutsites.append(use_chunk_cutsites)

        pos, counts = np.unique(np.concatenate(total_cutsites), return_counts=True)
        return pos, counts.astype("uint32")

    def _get_barcode_category_sites_count(self, barcodes_category):
        """
        Get the cutsite counts for multiple barcodes categories provided as a pandas series.

        Parameters
        ----------
        barcodes_category : pd.Series
            A pandas series with barcode as index and category as values.

        Returns
        -------
        dict
            A dictionary with category as keys and a tuple of numpy arrays (position, count) as values.
        """
        # convert barcode idx to category
        idx_to_category = barcodes_category.copy()
        idx_to_category.index = idx_to_category.index.map(self.barcode_to_idx)
        idx_to_category = idx_to_category.astype("str").astype("category")

        use_idx = self.get_barcodes_idx(barcodes_category.index)
        if use_idx.size == 0:
            return {cat: (np.array([]), np.array([])) for cat in barcodes_category.astype(str).unique()}

        # cutsite chunks
        site_chunks = self["cutsite"].chunks[0]
        chunk_boarders = [0] + list(np.cumsum(site_chunks))

        total_cutsites = {cat: [] for cat in barcodes_category.astype(str).unique()}
        for _i in range(len(site_chunks)):
            chunk_start = chunk_boarders[_i]
            chunk_end = chunk_boarders[_i + 1]

            # first, select barcode idx bool array
            use_chunk_sites = np.isin(self["cutsite"][chunk_start:chunk_end, 0], use_idx)
            # second, load actual cutsite global positions with the bool array
            use_chunk_cutsites = self["cutsite"][chunk_start:chunk_end, :][use_chunk_sites].to_pandas()
            # third, group by category and append to the total_cutsites
            for group, group_global_pos in use_chunk_cutsites["global_pos"].groupby(
                use_chunk_cutsites["barcode"].map(idx_to_category)
            ):
                total_cutsites[group].append(group_global_pos.values)

        # convert the list of arrays to a single array and count the unique values
        total_cutsites = {
            cat: np.unique(np.concatenate(sites), return_counts=True) for cat, sites in total_cutsites.items()
        }
        return total_cutsites

    def _sites_count_to_coverage_zarr(self, pos, counts, zarr_path, chunk_size):
        """Convert sites count to coverage."""
        coverage = np.zeros(self.genome_total_length, dtype=np.uint32)
        coverage[pos] = counts

        # cut the coverage to the chromosome lengths and store in zarr
        root = zarr.group(store=zarr_path, overwrite=True)

        for chrom, (
            global_start,
            global_end,
            _,
        ) in self.chrom_offsets.iterrows():
            root.create_dataset(
                f"chrs/{chrom}",
                data=coverage[global_start:global_end],
                chunks=chunk_size,
                compressor=self._zarr_compressor,
            )
        return

    def _category_sites_count_to_coverage_zarr(self, total_cutsites, zarr_path, chunk_size=20000000):
        """
        Save the cutsite counts for multiple categories to zarr file.

        The data will be separated by chromosome, and each chromosome will be saved as a separate dataset in the zarr file.
        For each chromosome, the data will be a category-by-position matrix, chunked by (1, chunk_size).
        """
        # cut the coverage to the chromosome lengths and store in zarr
        root = zarr.group(store=zarr_path, overwrite=True)

        n_categories = len(total_cutsites)
        # create empty datasets for each chromosome
        for chrom, size in self.chrom_offsets["size"].items():
            root.create_dataset(
                f"chrs/{chrom}",
                shape=(n_categories, size),
                chunks=(1, chunk_size),
                dtype=np.uint32,
                compressor=self._zarr_compressor,
            )

        # save each category to the zarr file
        cat_to_idx = {}
        for idx, (cat, (pos, counts)) in enumerate(total_cutsites.items()):
            cat_to_idx[cat] = idx
            coverage = np.zeros(self.genome_total_length, dtype=np.uint32)
            coverage[pos] = counts
            for chrom, (
                global_start,
                global_end,
                _,
            ) in self.chrom_offsets.iterrows():
                root[f"chrs/{chrom}"][idx, :] = coverage[global_start:global_end]
        return

    def _sites_count_to_bigwig(self, pos, counts, bigwig_path, resolution_bp=1):
        """Convert sites count to bigwig."""
        coverage = np.zeros(self.genome_total_length, dtype=np.uint32)
        coverage[pos] = counts

        with pyBigWig.open(bigwig_path, "w") as bw:
            header = list(self.chrom_offsets["size"].items())
            bw.addHeader(header)
            for chrom, (
                global_start,
                global_end,
                _,
            ) in self.chrom_offsets.iterrows():
                chrom_coverage = coverage[global_start:global_end]
                chrom_pos = np.where(chrom_coverage > 0)[0]
                chrom_counts = chrom_coverage[chrom_pos]
                if resolution_bp > 1:
                    resolution_bp = int(resolution_bp)
                    pos_counts = pd.Series(chrom_counts, index=chrom_pos // resolution_bp * resolution_bp)
                    pos_counts = pos_counts.groupby(pos_counts.index).sum()
                    chrom_pos = pos_counts.index.values
                    chrom_counts = pos_counts.values

                bw.addEntries(
                    chrom,
                    chrom_pos,
                    values=chrom_counts.astype("float32"),
                    span=resolution_bp,
                )
        return

    def dump_barcodes_coverage(self, barcodes, zarr_path, chunk_size=50000000):
        """Get the coverage for the barcodes and save to zarr file."""
        pos, counts = self._get_barcodes_sites_count(barcodes)
        self._sites_count_to_coverage_zarr(pos, counts, zarr_path, chunk_size)
        return

    def dump_barcodes_category_coverage(self, barcodes_category, zarr_path, chunk_size=50000000):
        """Get the coverage for the barcodes and save to zarr file."""
        total_cutsites = self._get_barcode_category_sites_count(barcodes_category)
        self._category_sites_count_to_coverage_zarr(total_cutsites, zarr_path, chunk_size)
        return

    def dump_barcodes_bigwig(self, barcodes, bigwig_path, resolution_bp=1):
        """Get the coverage for the barcodes and save to bigwig file."""
        pos, counts = self._get_barcodes_sites_count(barcodes)
        self._sites_count_to_bigwig(pos, counts, bigwig_path, resolution_bp)
        return
